from b22ao.aosystem import AOSystem
from b22ao.message import Message, State


class BaseOperation:

    """
    Base class for any adaptive optics routine.
    Children implement #start and #halt, and can #deform the mirrors and #capture data from the camera at their leisure!
    """

    def __init__(self):

        self.ao = AOSystem()
        self.config = None
        self.listener = None

    def attach_listener(self, listener):
        self.listener = listener

    def load_config(self, config):
        if config:
            import json
            with open(config, 'r') as doc:
                self.config = json.load(doc)

    def select_dm(self, dm):
        """
        If your entire operation involves a single mirror, you can specify it here

        e.g. passed in as config:

        self.select_dm(self.config['mirror'])  # where "mirror": 2 in config file
        # later:
        self.deform(mask)  # no need to specify mirror in subsequent calls
        """
        self.ao.select_dm(dm)

    def deform(self, mask, mirror=None):
        """
        Send a mask to specified mirror (1 or 2).

        If a mirror was previously selected with #select_dm
        and no mirror is specified here, the mask is applied
        to the mirror selected earlier.

        Bear in mind that the deformable mirrors in B22 *do not* support readback,
        so there is no guarantee that the mask is fully applied by the time this
        function returns; it may be necessary to wait for an empirically-determined
        time before characterising the system.
        """
        self.ao.deform(mask, mirror)

    def capture(self):
        """
        Returns a single detector frame
        """
        return self.ao.capture()

    def run(self):
        """
        Do not override. Implement #start instead
        """
        if self.listener:
            self.listener.notify(Message(self, State.Started))
        self.start()
        if self.listener:
            self.listener.notify(Message(self, State.Finished))

    def abort(self):
        """
        Do not override. Implement #halt instead
        """
        self.stop()
        if self.listener:
            self.listener.notify(Message(self, State.Failed, "Aborted"))

    def start(self):
        """
        Starts the operation

        If a JSON configuration file was specified,
        implementations can now access the dictionary self.config
        """
        raise NotImplementedError

    def stop(self):
        """
        Stops the operation
        """
        raise NotImplementedError



from epics import caget, caput, camonitor
from b22ao.pvs import *
from enum import Enum
from threading import Event
import numpy


class AOSystem:

    def __init__(self):
        self.cam = AreaDetector(AD_PVBASE)
        self.dm = None  # DM of choice

    def select_dm(self, dm):
        self.dm = DeformableMirror(dm)

    def deform(self, mask, dm=None):
        try:
            DeformableMirror(dm).deform(mask)
        except ValueError:  # no dm provided
            try:
                self.dm.deform(mask)
            except AttributeError:
                raise NameError("Call #select_dm(dm) first or specify a DM in #deform(mask, dm)")

    def capture(self):
        return self.cam.acquire()

    def get_metadata(self):
        return {'cam': self.cam.get_metadata(), 'dm': self.dm.get_metadata()}


class DeformableMirror(Enum):

    DM1 = 1
    DM2 = 2

    def deform(self, mask):
        for actuator in range(len(mask)):
            caput(self._get_pv_base() + DM_ACTUATOR_PREFIX + str(actuator) + DM_ACTUATOR_SETPOINT, mask[actuator])
        caput(self._get_pv_base() + DM_APPLY_MASK, 1)

    def _get_pv_base(self):
        if self is DeformableMirror.DM1:
            return standardise_pv(DM1_PVBASE)
        elif self is DeformableMirror.DM2:
            return standardise_pv(DM2_PVBASE)

    def get_metadata(self):
        return {'mirror': self}


class AreaDetector:

    def __init__(self, pv_base):
        pv_base = standardise_pv(pv_base)
        self.data_ready = Event()
        self.current_frame = None
        self.frame_counter_pv = pv_base + AD_ARRAY_COUNTER
        self.acquire_pv = pv_base + AD_ACQUIRE
        self.data_pv = pv_base + AD_ARRAY_DATA
        self.dim_x_pv = pv_base + AD_DIM_X
        self.dim_y_pv = pv_base + AD_DIM_Y
        self.gain_pv = pv_base + AD_GAIN

        caput(pv_base + AD_IMAGE_MODE, AD_IMAGE_MODE_SINGLE)

        camonitor(self.frame_counter_pv, callback=self.frame_counter_callback)

    def acquire(self):
        self.current_frame = caget(self.frame_counter_pv)
        caput(self.acquire_pv, "Acquire")
        self.data_ready.wait()
        data = caget(self.data_pv)
        self.data_ready.clear()
        data = numpy.reshape(data, (caget(self.dim_y_pv), caget(self.dim_x_pv)))
        return data

    def frame_counter_callback(self, **kwargs):
        if kwargs['value'] == self.current_frame + 1:
            self.data_ready.set()

    def get_metadata(self):
        metadata = dict()
        metadata['gain'] = caget(self.gain_pv)
        return metadata


def standardise_pv(pv):
    return pv if pv.endswith(':') else pv + ':'




AD_PVBASE = "BL22B-DI-CAM-10"  # Area Detector

DM1_PVBASE = "BL22B-OP-ALPAO-01"  # Deformable mirror 1
DM2_PVBASE = "BL22B-OP-ALPAO-02"  # Deformable mirror 2


AD_ARRAY_COUNTER = "CAM:ArrayCounter_RBV"  # Frame counter
AD_ACQUIRE = "CAM:Acquire"  # Acquire button - write with 'Acquire'
AD_GAIN = "CAM:Gain_RBV"  # Gain set on Windows app
AD_ARRAY_DATA = "ARR:ArrayData"  # Data
AD_DIM_X = "ARR:ArraySize0_RBV"
AD_DIM_Y = "ARR:ArraySize1_RBV"
AD_IMAGE_MODE = "CAM:ImageMode"
AD_IMAGE_MODE_SINGLE = "Single"

DM_ACTUATOR_PREFIX = "ACT"  # individual actuator pv: pv base + ACT + actuator number
DM_ACTUATOR_SETPOINT = ":SP"
DM_APPLY_MASK = "CP_ST_TO_ACT.PROC"



import enum


class State(enum.Enum):
    Started = 1
    Finished = 2
    Failed = 3


class Message:
    def __init__(self, source, state, msg=None):
        self.source = source
        self.state = state
        self.msg = msg
